#ifndef vtk_m_cont_PointLocatorSelfJoin_h
#define vtk_m_cont_PointLocatorSelfJoin_h

#include <vtkm/worklet/WorkletMapField.h>

#include <vtkm/cont/Algorithm.h>
#include <vtkm/cont/PointLocator.h>
#include <vtkm/cont/TryExecute.h>
#include <vtkm/cont/ArrayHandleCounting.h>
#include <vtkm/exec/PointLocatorSelfJoin.h>
#include <vtkm/worklet/DispatcherMapField.h>

namespace vtkm
{

namespace cont
{

namespace detail
{

class PointToCellId : public vtkm::worklet::WorkletMapField
{
public:
  using ControlSignature = void(FieldIn point,
                                FieldOut cellId);
  using ExecutionSignature = void(_1, WorkIndex, _2);

  template <typename Point>
  VTKM_EXEC void operator()(const Point& point,
                            const vtkm::Id& index,
                            vtkm::Id& cellId) const
  {
    vtkm::Vec<vtkm::FloatDefault, 3> relative;
    // Getting the absolute value below, after observing a case were small
    // floating point problems.
    relative[0] = static_cast<vtkm::FloatDefault>(vtkm::Abs(point[0] - Bounds.X.Min) * Scale[0]);
    relative[1] = static_cast<vtkm::FloatDefault>(vtkm::Abs(point[1] - Bounds.Y.Min) * Scale[1]);
    relative[2] = static_cast<vtkm::FloatDefault>(vtkm::Abs(point[2] - Bounds.Z.Min) * Scale[2]);
    vtkm::Id3 logicalCell;
    logicalCell[0] = static_cast<vtkm::IdComponent>(vtkm::Floor(relative[0]));
    logicalCell[1] = static_cast<vtkm::IdComponent>(vtkm::Floor(relative[1]));
    logicalCell[2] = static_cast<vtkm::IdComponent>(vtkm::Floor(relative[2]));
    cellId = logicalCell[2] * PlaneSize + logicalCell[1] * RowSize + logicalCell[0];
  }

  VTKM_CONT
  PointToCellId(vtkm::Bounds& bounds,
                vtkm::Id3& dims)
    : Bounds(bounds)
    , Dims(dims)
  {
    Scale[0]
      = static_cast<vtkm::FloatDefault>((dims[0] - 1.0l) / (bounds.X.Max - bounds.X.Min));
    Scale[1]
      = static_cast<vtkm::FloatDefault>((dims[1] - 1.0l) / (bounds.Y.Max - bounds.Y.Min));
    Scale[2]
      = static_cast<vtkm::FloatDefault>((dims[2] - 1.0l) / (bounds.Z.Max - bounds.Z.Min));
    PlaneSize = dims[0] * dims[1];
    RowSize = dims[0];
  }

private:
  vtkm::Bounds Bounds;
  vtkm::Vec<vtkm::FloatDefault, 3> Scale;
  vtkm::Id3 Dims;
  vtkm::Id PlaneSize;
  vtkm::Id RowSize;
};

} // namespace detail

class PointLocatorSelfJoin : public vtkm::cont::PointLocator
{
private:
  using IdHandle = vtkm::cont::ArrayHandle<vtkm::Id>;
  using Point = vtkm::Vec<vtkm::FloatDefault, 3>;
  using PointHandle = vtkm::cont::ArrayHandle<Point>;

public:
  PointLocatorSelfJoin(vtkm::Bounds& bounds,
                       vtkm::Id3& dims,
                       vtkm::FloatDefault& threshold)
    : PointLocator()
    , Bounds(bounds)
    , Dims(dims)
    , Threshold(threshold)
  {
    vtkm::FloatDefault xSpacing = static_cast<vtkm::FloatDefault>(bounds.X.Length() / dims[0]);
    vtkm::FloatDefault ySpacing = static_cast<vtkm::FloatDefault>(bounds.Y.Length() / dims[1]);
    vtkm::FloatDefault zSpacing = static_cast<vtkm::FloatDefault>(bounds.Z.Length() / dims[2]);
    this->LookUpLimit[0] = vtkm::Ceil(threshold / xSpacing);
    this->LookUpLimit[1] = vtkm::Ceil(threshold / ySpacing);
    this->LookUpLimit[2] = vtkm::Ceil(threshold / zSpacing);
  }

  struct PrepareForExecutionFunctor
  {
    template <typename DeviceAdapter>
    VTKM_CONT bool operator()(DeviceAdapter,
                              const vtkm::cont::PointLocatorSelfJoin& self,
                              HandleType& handle) const
    {
      vtkm::exec::PointLocatorSelfJoin<DeviceAdapter>* execObject =
        new vtkm::exec::PointLocatorSelfJoin<DeviceAdapter>(
          self.Threshold,
          self.Dims,
          self.LookUpLimit,
          self.Points,
          self.CellIds,
          self.GroupedPoints,
          self.LowerIndices,
          self.UpperIndices);
      handle.Reset(execObject);
      return true;
    }
  };

  VTKM_CONT
  const HandleType PrepareForExecutionImpl(vtkm::cont::DeviceAdapterId deviceId) const override
  {
    const bool success =
      vtkm::cont::TryExecuteOnDevice(deviceId, PrepareForExecutionFunctor(), *this, ExecHandle);
    if (!success)
    {
      throwFailedRuntimeDeviceTransfer("PointLocatorUniformGrid", deviceId);
    }
    return ExecHandle;
  }

  void Build() override
  {
      // Group together all the points in the same cell.
      // And use them to find the points that are close
      // by for the same cell.
      // 1. From the Coord System Group all points together.
      // 2. Points are to be considered with indices.
      // 3. Sort all point indices according to the cells they belong.
      //    Here we're only dealing with point Ids, and not actual points.
      //    Now the actual distances will be calculated when we actually are
      //    in the method FindNearestNeighbor.
      vtkm::cont::Algorithm::Copy(this->GetCoordinates().GetData(), this->Points);
      IdHandle cellIds;
      vtkm::cont::ArrayHandleIndex indices(this->Points.GetNumberOfValues());
      vtkm::cont::Algorithm::Copy(indices, this->GroupedPoints);

      // Gather cell Ids for all the points provided in the input.
      detail::PointToCellId worklet(this->Bounds, this->Dims);
      vtkm::worklet::DispatcherMapField<detail::PointToCellId> dispatcher(worklet);
      dispatcher.Invoke(this->Points, cellIds);
      // For the need to know which point maps to which cell.
      // This needs to be done before this is used for sorting.
      vtkm::cont::Algorithm::Copy(cellIds, this->CellIds);

      // Now we have grouped together points that occur in the same cell.
      vtkm::cont::Algorithm::SortByKey(cellIds, this->GroupedPoints);
      // Get the lookup start, and lookup end for each cell.
      vtkm::cont::ArrayHandleCounting<vtkm::Id>
        forBounds(0, 1, this->Dims[0] * this->Dims[1] * this->Dims[2]);
      vtkm::cont::Algorithm::LowerBounds(cellIds, forBounds, this->LowerIndices);
      vtkm::cont::Algorithm::UpperBounds(cellIds, forBounds, this->UpperIndices);
  }

private:
  vtkm::Bounds Bounds;
  vtkm::Id3 Dims;
  vtkm::FloatDefault Threshold;
  vtkm::Id3 LookUpLimit;
  PointHandle Points;
  IdHandle CellIds;
  IdHandle GroupedPoints;
  IdHandle LowerIndices;
  IdHandle UpperIndices;
  mutable HandleType ExecHandle;
};

} //namespace cont
} //namespace exec

#endif // vtk_m_cont_PointLocatorSelfJoin_h
