#ifndef vtk_m_exec_PointLocatorSelfJoin_h
#define vtk_m_exec_PointLocatorSelfJoin_h

#include <vtkm/Types.h>
#include <vtkm/VectorAnalysis.h>

#include <vtkm/cont/ArrayHandle.h>

#include <vtkm/exec/PointLocator.h>

namespace vtkm
{

namespace exec
{

template <typename DeviceAdapter>
class PointLocatorSelfJoin : public vtkm::exec::PointLocator
{
private:
  using IdHandle = typename vtkm::cont::ArrayHandle<vtkm::Id>;
  using IdHandleExec
    = typename vtkm::cont::ArrayHandle<vtkm::Id>::template ExecutionTypes<DeviceAdapter>::PortalConst;
  using Point = vtkm::Vec<vtkm::FloatDefault, 3>;
  using PointHandle = typename vtkm::cont::ArrayHandle<Point>;
  using PointHandleExec
    = typename vtkm::cont::ArrayHandle<Point>::template ExecutionTypes<DeviceAdapter>::PortalConst;

public:
  PointLocatorSelfJoin() = default;

  PointLocatorSelfJoin(const vtkm::FloatDefault& threshold,
                       const vtkm::Id3& dims,
                       const vtkm::Id3& lookUpLimit,
                       const PointHandle& points,
                       const IdHandle& cellIds,
                       const IdHandle& groupedPoints,
                       const IdHandle& lowerIndices,
                       const IdHandle& upperIndices)
    : Threshold(threshold)
    , Dims(dims)
    , LookUpLimit(lookUpLimit)
  {
    this->Points = points.PrepareForInput(DeviceAdapter());
    this->CellIds = cellIds.PrepareForInput(DeviceAdapter());
    this->GroupedPoints = groupedPoints.PrepareForInput(DeviceAdapter());
    this->LowerIndices = lowerIndices.PrepareForInput(DeviceAdapter());
    this->UpperIndices = upperIndices.PrepareForInput(DeviceAdapter());
    this->PlaneSize = dims[0] * dims[1];
    this->RowSize = dims[0];
  }

  VTKM_EXEC void FindNearestNeighbor(
    const Point& vtkmNotUsed(queryPoint),
    vtkm::Id& vtkmNotUsed(nearestNeighborId),
    vtkm::FloatDefault& vtkmNotUsed(distance2)) const override
  {}

  VTKM_EXEC
  vtkm::Id GetCellId(vtkm::Id x, vtkm::Id y, vtkm::Id z) const
  {
    return (z * this->PlaneSize + y * this->RowSize + x);
  }

  VTKM_EXEC
  void GetLogicalCellId(vtkm::Id cellId, vtkm::Id3& logical) const
  {
    logical[0] = cellId%(this->Dims[0]-1);
    logical[1] = (cellId/(this->Dims[0]-1))%(this->Dims[1]-1);
    logical[2] = cellId/((this->Dims[0]-1)*(this->Dims[1]-1));
  }

  VTKM_EXEC
  void FindNearestNeighbor(const vtkm::Id& pointId,
                           vtkm::Id& neighbor,
                           vtkm::FloatDefault& distance) const override
  {
    neighbor = -1;
    distance = vtkm::Infinity<vtkm::FloatDefault>();
    Point x = Points.Get(pointId);
    vtkm::Id cellId = CellIds.Get(pointId);
    //std::cout << pointId << " : " << cellId << std::endl;
    vtkm::Id start = LowerIndices.Get(cellId);
    vtkm::Id end = UpperIndices.Get(cellId);
    for(vtkm::Id index = start; index < end; index++)
    {
      vtkm::Id nnInd = GroupedPoints.Get(index);
      // Since this is self join, we need to skip comparing a point
      // with itself, as this will always be the minimum case.
      if(pointId == nnInd)
        continue;
      Point y = Points.Get(nnInd);
      vtkm::FloatDefault y_distance = vtkm::MagnitudeSquared(x - y);
      if(y_distance == distance)
      {
        neighbor = vtkm::Min(neighbor, nnInd);
      }
      else if(y_distance < distance)
      {
        neighbor = nnInd;
        distance = y_distance;
      }
    }
  }

private:
  vtkm::FloatDefault Threshold;
  vtkm::Id3 Dims, LookUpLimit;
  PointHandleExec Points;
  IdHandleExec CellIds, GroupedPoints, LowerIndices, UpperIndices;
  vtkm::Id PlaneSize;
  vtkm::Id RowSize;
};

} // namespace cont
} // namespace vtkm

#endif //vtk_m_exec_PointLocatorSelfJoin_h
